/*
Kart Rush: Race The Clock
Yet another Modified version of Latius' Countdown mode, by Latius and Professor Renderer

Credits to the following people:

Fav for the old Hat in Time Timer sprites
Afrorule for their Start Boost Helper which has been modified
Latius for the custom-made 7-segment display used for the Timer
Latius again for extracting the PA sounds from A Hat In Time
Professor Renderer for GP Mode music, Death Wish music and beta testing
Professor Renderer again for extracting "Get Ready," "Go," and "Finish" sounds from A Hat in Time
*/

local MODVERSION = "1.1"

local NOOBMOD = 50
local EASIESTMOD = 25
local EASYMOD = 10
local HARDMOD = -5
local EXPERTMOD = -10
local MASTERMOD = -20
local INSANEMOD = -30

//Difficulty is currently global, but may set it to vary per player
local difficulty = CV_RegisterVar({"kr_diff", "Normal", CV_NETVAR|CV_SHOWMODIF, {Noob=NOOBMOD, Easiest=EASIESTMOD, Easy=EASYMOD, Normal=0, Hard=HARDMOD, Expert=EXPERTMOD, Master=MASTERMOD, Insane=INSANEMOD}})
local forceUpdate = CV_RegisterVar({"kr_force", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local adaptiveDiff = CV_RegisterVar({"kr_adapt", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local playdiff = difficulty.value
local clockrace = true
local FLASHTIME = 5 * TICRATE / 2
local TIMEOVERLENGTH = 30 * TICRATE
local TIMEFORTHREE = 26 + (TICRATE * 3)
local TIMEFORGO = 26 + (TICRATE * 6)
local BASETIME = 150 //The default time bonus per map

local MILESPERHOUR = 1
local KILOMETERSPERHOUR = 2
local NAUTICALMILESPERHOUR = 3
local METERSPERSECOND = 4
local FOOTPERSECOND = 5
local FRACUNITSPERTIC = 6

local NUMSKINCOLORS = 101

local dspeed = CV_RegisterVar({"kr_daytona", "Off", CV_SHOWMODIF, CV_OnOff})
local tenthspeedometer = CV_RegisterVar({"kr_daytonatenth", "On", CV_SHOWMODIF, CV_OnOff})
local customspeedometer = CV_RegisterVar({"kr_daytonaspeed", "Miles", CV_SHOWMODIF, {Miles=MILESPERHOUR, Kilometers=KILOMETERSPERHOUR, Meters=METERSPERSECOND, Fracunits=FRACUNITSPERTIC, Knots=NAUTICALMILESPERHOUR, Feet=FOOTPERSECOND}})

local deathwish = CV_RegisterVar({"kr_deathwish", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local deathwishdiff = CV_RegisterVar({"kr_dwdiff", "Normal", CV_NETVAR|CV_SHOWMODIF, {Insane = 0, Expert = 5, Hard = 7, Normal = 10, Easy = 15, Easiest = 20}})
local dwon = deathwish.value
local dwdiff = deathwishdiff.value
local dwsoul = CV_RegisterVar({"kr_dwsoul", "Off", CV_SHOWMODIF, CV_OnOff})

local comments = CV_RegisterVar({"kr_comments", "On", CV_SHOWMODIF, CV_OnOff})

local gpmode = CV_RegisterVar({"kr_gp", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local gpmusic = CV_RegisterVar({"kr_gpmusic", "Off", CV_SHOWMODIF, CV_OnOff})
local gprift = CV_RegisterVar({"kr_gprift", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})

local DEFAULTGP = 1
local LEGACYGP = 2
local ROOKIEGP = 3
local CHALLENGEGP = 4
local ZONEGP = 5

local gptypevar = CV_RegisterVar({"kr_gptype", "Default", CV_SHOWMODIF|CV_NETVAR, {Default=DEFAULTGP, Legacy=LEGACYGP, Rookie=ROOKIEGP, Challenge=CHALLENGEGP, Zone=ZONEGP}})
local gptype = DEFAULTGP

local driftbonus = CV_RegisterVar({"kr_driftplus", "Off", CV_SHOWMODIF, CV_OnOff})

COM_AddCommand("kr_welcome", function(player)
	CONS_Printf(player, "\x82".."Welcome to Kart Rush: Race the Clock, "..MODVERSION)
	CONS_Printf(player, "\x82")
	CONS_Printf(player, "\x82".."Commands:")
	CONS_Printf(player, "\x82".."kr_welcome, kr_help, kr_gphelp")
	CONS_Printf(player, "\x82")
	CONS_Printf(player, "\x82".."CVars:")
	CONS_Printf(player, "\x82".."kr_daytona, kr_daytonatenth, kr_daytonaspeed, kr_dwsoul, kr_gpmusic, kr_comments")
	CONS_Printf(player, "\x82")
	CONS_Printf(player, "\x82".."NetVars:")
	CONS_Printf(player, "\x82".."kr_diff, kr_force, kr_adapt, kr_deathwish, kr_dwdiff, kr_gp, kr_gprift, kr_gptype, kr_driftplus")
end)

COM_AddCommand("kr_help", function(player, command)
	if command == nil then
		CONS_Printf(player, "\x82".."Usage: kr_help <variable>")
	else
		if (command == "kr_welcome") or (command == "welcome") then
			CONS_Printf(player, "\x82".."Shows the welcome message again.")
		elseif (command == "kr_help") or (command == "help") then
			CONS_Printf(player, "\x82".."ERROR: can't show help for command "..command.." - You know how to use it already, silly!")
		elseif (command == "kr_gphelp") or (command == "gphelp") then
			CONS_Printf(player, "\x82".."Shows the current GP settings")
		elseif (command == "kr_daytona") or (command == "daytona") then
			CONS_Printf(player, "\x82".."Enables or disables the Daytona-style speedometer. Default off.")
			if dspeed.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_daytonatenth") or (command == "daytonatenth") then
			CONS_Printf(player, "\x82".."Enables or disables the tenths place on the Daytona-style speedometer. Default on.")
			if tenthspeedometer.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_daytonaspeed") or (command == "daytonaspeed") then
			CONS_Printf(player, "\x82".."Sets speedometer type on the Daytona-style speedometer. Default miles.")
			if customspeedometer.value == MILESPERHOUR then
				CONS_Printf(player, "\x82".."This is currently set to MILES.")
			elseif customspeedometer.value == KILOMETERSPERHOUR then
				CONS_Printf(player, "\x82".."This is currently set to KILOMETERS.")
			elseif customspeedometer.value == METERSPERHOUR then
				CONS_Printf(player, "\x82".."This is currently set to METERS.")
			elseif customspeedometer.value == FRACUNITSPERTIC then
				CONS_Printf(player, "\x82".."This is currently set to FRACUNITS.")
			elseif customspeedometer.value == NAUTICALMILESPERHOUR then
				CONS_Printf(player, "\x82".."This is currently set to KNOTS.")
			else
				CONS_Printf(player, "\x82".."This is currently set to FEET.")
			end
		elseif (command == "kr_dwsoul") or (command == "dwsoul") then
			CONS_Printf(player, "\x82".."Enables or disables the Death Wish soul timer. Default off.")
			if dwsoul.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_gpmusic") or (command == "gpmusic") then
			CONS_Printf(player, "\x82".."Enables or disables A Hat in Time music during GP Mode. Default off.")
			if gpmusic.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_diff") or (command == "diff") then
			CONS_Printf(player, "\x82".."Changes difficulty, which affects how much time you start and how much you gain on each map clear. Host only, default normal.")
			CONS_Printf(player, "\x82".."Possible values: Noob, Easiest, Easy, Normal, Hard, Expert or Master.")
			if difficulty.value == NOOBMOD then
				CONS_Printf(player, "\x86".."Difficulty is currently set to NOOB. (150% Time Limit, 50% Score)")
			elseif difficulty.value == EASIESTMOD then
				CONS_Printf(player, "\x84".."Difficulty is currently set to EASIEST. (125% Time Limit, 75% Score)")
			elseif difficulty.value == EASYMOD then
				CONS_Printf(player, "\x84".."Difficulty is currently set to EASY. (110% Time Limit, 90% Score)")
			elseif difficulty.value == 0 then
				CONS_Printf(player, "\x82".."Difficulty is currently set to NORMAL. (Default Time Limit and Score)")
			elseif difficulty.value == HARDMOD then
				CONS_Printf(player, "\x87".."Difficulty is currently set to HARD. (95% Time Limit, 105% Score)")
			elseif difficulty.value == EXPERTMOD then
				CONS_Printf(player, "\x87".."Difficulty is currently set to EXPERT. (90% Time Limit, 120% Score)")
			elseif difficulty.value == MASTERMOD then
				CONS_Printf(player, "\x85".."Difficulty is currently set to MASTER. (80% Time Limit, 180% Score)")
			else
				CONS_Printf(player, "\x85".."Difficulty is currently set to INSANE. (70% Time Limit, 280% Score)")
			end
		elseif (command == "kr_force") or (command == "force") then
			CONS_Printf(player, "\x82".."Enables or disables difficulty forcing. When enabled, difficulty will be forced on the next map. Host only, default off.")
			if forceUpdate.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_adapt") or (command == "adapt") then
			CONS_Printf(player, "\x82".."Enables or disables adaptive difficulty. When enabled, difficulty will change based on how well you do each map. Host only, default off.")
			if adaptiveDiff.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_gp") or (command == "gp") then
			CONS_Printf(player, "\x82".."Enables or disables GP Mode, a mode that takes you through random courses with increasing time difficulty. Host only, default off.")
			if gpmode.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_deathwish") or (command == "deathwish") then
			CONS_Printf(player, "\x82".."Enables or disables ".."\x81".."Death Wish".."\x82".." Mode, a mode that gives you 999 seconds (plus a fraction of the normal time you'd normally get for each map after the first) to complete as many courses as possible.")
			CONS_Printf(player, "\x82".."Bumping into walls or falling off the map loses time. Host only, default off.")
			if deathwish.value == 1 then
				CONS_Printf(player, "\x81".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_dwdiff") or (command == "dwdiff") then
			CONS_Printf(player, "\x82".."Sets the difficulty of Death Wish Mode, bewteen Easiest, Easy, Normal, Hard, Expert or Insane.")
			CONS_Printf(player, "\x82".."Higher difficulties give less time for each map after the first. Host only, default Normal.")
			if dwdiff == 0 then
				CONS_Printf(player, "\x82".."Difficulty set to INSANE (NO Time Bonus per map)")
			elseif dwdiff == 5 then
				CONS_Printf(player, "\x82".."Difficulty set to EXPERT (25% Time Bonus per map)")
			elseif dwdiff == 7 then
				CONS_Printf(player, "\x82".."Difficulty set to HARD (35% Time Bonus per map)")
			elseif dwdiff == 10 then
				CONS_Printf(player, "\x82".."Difficulty set to NORMAL (50% Time Bonus per map)")
			elseif dwdiff == 15 then
				CONS_Printf(player, "\x82".."Difficulty set to EASY (75% Time Bonus per map)")
			else
				CONS_Printf(player, "\x82".."Difficulty set to EASIEST (Full Time Bonus per map)")
			end
		elseif (command == "kr_gprift") or (command == "gprift") then
			CONS_Printf(player, "\x82".."Enables or disables Rift Collapse effects when playing Encore maps during GP Mode.")
			CONS_Printf(player, "\x85".."WARNING: IF YOU ARE PRONE TO SEIZURES, DO NOT TURN ON THIS COMMAND. ".."\x85".."Default off.")
			if dwsoul.value == 1 then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_gptype") or (command == "gptype") then
			CONS_Printf(player, "\x82".."Sets GP type. These change how many map clears until the next difficulty in GP Mode. Default 'Default'")
			CONS_Printf(player, "\x82".."Possible values: Default, Legacy, Rookie, Challenge or Zone")
			if gptypevar.value == DEFAULTGP then
				CONS_Printf(player, "\x82".."The current GP type is DEFAULT GP, the normal setting.")
			elseif gptypevar.value == LEGACYGP then
				CONS_Printf(player, "\x82".."The current GP type is LEGACY GP, used in beta testing.")
			elseif gptypevar.value == ROOKIEGP then
				CONS_Printf(player, "\x82".."The current GP type is ROOKIE GP, for beginners.")
			elseif gptypevar.value == CHALLENGEGP then
				CONS_Printf(player, "\x82".."The current GP type is CHALLENGE GP, for advanced players.")
			else
				CONS_Printf(player, "\x82".."The current GP type is ZONE GP, based on Wipeout's Zone mode.")
			end	
		elseif (command == "kr_driftplus") or (command == "driftplus") then
			CONS_Printf(player, "\x82".."Enables or disables the DTF (Drift Time Freeze) Function. If enabled, scores are cut by 50% (25% in Death Wish). Default Off")
			if  (driftbonus.value == 1) then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		elseif (command == "kr_comments") or (command == "comments") then
			CONS_Printf(player, "\x82".."Enables or disables 'commentarty' at the end of each lap about your pace and odds of survival. Default On")
			if  (comments.value == 1) then
				CONS_Printf(player, "\x84".."This is currently ENABLED.")
			else
				CONS_Printf(player, "\x85".."This is currently DISABLED.")
			end
		else
			CONS_Printf(player, "\x82".."ERROR: can't show help for command "..command.." (404 Command Not Found)")
		end
	end
end)

COM_AddCommand("kr_gphelp", function(player)
	local mode = gptypevar
	if gptypevar.value == DEFAULTGP then
		CONS_Printf(player, "\x82".."The current GP type is DEFAULT GP.")
		CONS_Printf(player, "\x84".."Race 1-2:   Easiest")
		CONS_Printf(player, "\x84".."Race 3-6:   Easy")
		CONS_Printf(player, "\x82".."Race 7-14:  Normal")
		CONS_Printf(player, "\x87".."Race 15-19: Hard")
		CONS_Printf(player, "\x87".."Race 20-25: Expert")
		CONS_Printf(player, "\x85".."Race 26-30: Master")
		CONS_Printf(player, "\x85".."Race 31+:   Insane")
	elseif gptypevar.value == LEGACYGP then
		CONS_Printf(player, "\x82".."The current GP type is LEGACY GP.")
		CONS_Printf(player, "\x84".."Race 1-2:   Easiest")
		CONS_Printf(player, "\x84".."Race 3-5:   Easy")
		CONS_Printf(player, "\x82".."Race 6-11:  Normal")
		CONS_Printf(player, "\x87".."Race 12-19: Hard")
		CONS_Printf(player, "\x87".."Race 20-31: Expert")
		CONS_Printf(player, "\x85".."Race 32-39: Master")
		CONS_Printf(player, "\x85".."Race 40+:   Insane")
	elseif gptypevar.value == ROOKIEGP then
		CONS_Printf(player, "\x82".."The current GP type is ROOKIE GP.")
		CONS_Printf(player, "\x84".."Race 1-4:   Easiest")
		CONS_Printf(player, "\x84".."Race 5-9:   Easy")
		CONS_Printf(player, "\x82".."Race 10-15: Normal")
		CONS_Printf(player, "\x87".."Race 16-22: Hard")
		CONS_Printf(player, "\x87".."Race 23-30: Expert")
		CONS_Printf(player, "\x85".."Race 31-34: Master")
		CONS_Printf(player, "\x85".."Race 35+:   Insane")
	elseif gptypevar.value == CHALLENGEGP then
		CONS_Printf(player, "\x82".."The current GP type is CHALLENGE GP.")
		CONS_Printf(player, "\x84".."Race 1-5:   Easy")
		CONS_Printf(player, "\x82".."Race 6-11:  Normal")
		CONS_Printf(player, "\x87".."Race 12-18: Hard")
		CONS_Printf(player, "\x87".."Race 19-26: Expert")
		CONS_Printf(player, "\x85".."Race 27-34: Master")
		CONS_Printf(player, "\x85".."Race 35+:   Insane")
	else
		CONS_Printf(player, "\x82".."The current GP type is ZONE GP.")
		CONS_Printf(player, "\x86".."Race 1:     NOOB")
		CONS_Printf(player, "\x84".."Race 2-4:   Easiest")
		CONS_Printf(player, "\x84".."Race 5-6:   Easy")
		CONS_Printf(player, "\x82".."Race 7-15:  Normal")
		CONS_Printf(player, "\x87".."Race 16-19: Hard")
		CONS_Printf(player, "\x87".."Race 20-26: Expert")
		CONS_Printf(player, "\x85".."Race 27-34: Master")
		CONS_Printf(player, "\x85".."Race 35+:   Insane")
	end	
end)

local NORMALPLACEHOLDER = 255 //Needed to fix a bug with GP modes

local gpsetup = {
	[DEFAULTGP] = {
		[1] = EASIESTMOD,
		[3] = EASYMOD,
		[7] = NORMALPLACEHOLDER,
		[15] = HARDMOD,
		[20] = EXPERTMOD,
		[26] = MASTERMOD,
		[31] = INSANEMOD
	},
	[LEGACYGP] = {
		[1] = EASIESTMOD,
		[3] = EASYMOD,
		[6] = NORMALPLACEHOLDER,
		[12] = HARDMOD,
		[20] = EXPERTMOD,
		[32] = MASTERMOD,
		[40] = INSANEMOD
	},
	[ROOKIEGP] = {
		[1] = EASIESTMOD,
		[5] = EASYMOD,
		[10] = NORMALPLACEHOLDER,
		[16] = HARDMOD,
		[23] = EXPERTMOD,
		[31] = MASTERMOD,
		[35] = INSANEMOD
	},
	[CHALLENGEGP] = {
		[1] = EASYMOD,
		[6] = NORMALPLACEHOLDER,
		[12] = HARDMOD,
		[19] = EXPERTMOD,
		[27] = MASTERMOD,
		[35] = INSANEMOD
	},
	[ZONEGP] = {
		[1] = NOOBMOD,
		[2] = EASIESTMOD,
		[5] = EASYMOD,
		[7] = NORMALPLACEHOLDER,
		[16] = HARDMOD,
		[20] = EXPERTMOD,
		[27] = MASTERMOD,
		[35] = INSANEMOD
	},
}

local function getgptier(p)
	local n = p.raceclock.clearedmaps + 1
	local tier = NOOBMOD
	for i = 1, min(40, n)	-- don't bother going above 40, there's nothing beyond that
		if gpsetup[gptype][i]
			tier = gpsetup[gptype][i]
		end
	end
	if tier == NORMALPLACEHOLDER then tier = 0 end
	return tier
end

local function gpmapsetup(p)
	if (gpmode.value ~= 1) then return end
	if p.raceclock.clearedmaps == 0 then
		COM_BufInsertText(server, "advancemap Random")
		gptype = gptypevar.value
	end
	playdiff = getgptier(p)
end

local clockattacker = 0 //The player number of the current Clock Attacker
local endtimer = 0
local ktching = false

addHook("NetVars", function(net)
	clockattacker = net(clockattacker)
	endtimer = net(endtimer)
	ktching = net(ktching)
	playdiff = net(playdiff)
	dwon = net(dwon)
end)

local flash = {
	[0]	= 655360, //Fully transparent
	[1] = V_90TRANS,
	[2] = V_80TRANS,
	[3] = V_70TRANS,
	[4] = V_60TRANS,
	[5] = V_50TRANS,
	[6] = V_40TRANS,
	[7] = V_30TRANS,
	[8] = V_20TRANS,
	[9] = V_10TRANS,
	[10] = 0 //Fully opaque
}

//The base Normal Times for each map
local maptimes = setmetatable({
	[1] = 120,
	[2] = 165,
	[3] = 165,
	[4] = 120,
	[5] = 135,
	[6] = 150,
	[7] = 150,
	[8] = 120,
	[9] = 180,
	[10] = 150,
	[11] = 120,
	[12] = 135,
	[13] = 135,
	[14] = 120,
	[15] = 120,
	[16] = 165,
	[17] = 180,
	[18] = 135,
	[19] = 180,
	[20] = 150,
	[21] = 150,
	[22] = 120,
	[23] = 150,
	[24] = 180,
	[25] = 150,
	[26] = 120,
	[27] = 165,
	[28] = 150,
	[29] = 180,
	[30] = 135,
	[31] = 165,
	[32] = 150,
	[33] = 150,
	[34] = 120,
	[35] = 165,
	[36] = 120,
	[37] = 135,
	[38] = 165,
	[39] = 150,
	[40] = 120,
	[41] = 105,
	[42] = 120,
	[43] = 120,
	[44] = 120,
	[45] = 135,
	[46] = 120,
	[47] = 135,
	[48] = 120,
	[49] = 105,
	[50] = 120,
	[51] = 105,
	[52] = 105,
	[53] = 90,
	[54] = 150,
	[55] = 90,
	[56] = 180,
	[57] = 150,
	[58] = 150,
	[59] = 180,
	[60] = 150,
	
	[352] = 225,
	[353] = 135,
	[354] = 120,
	[355] = 180,
	[356] = 195,
	[357] = 180,
	[358] = 135,
	[359] = 150,
	[360] = 165,
	[361] = 165,
	[362] = 150,
	[363] = 180,
	[364] = 150,
	[365] = 165,
	[369] = 120
	
}, {__index = function() return BASETIME end})

local trainPA = CV_RegisterVar({"rtc_pa", "On", CV_SHOWMODIF, CV_OnOff})

local sfxtemp

for sfxtemp = 0, 9 do
freeslot("sfx_tra00"..sfxtemp)
end
freeslot(
"sfx_tra010", "sfx_tra040", "sfx_tra080", "sfx_tra120", "sfx_trafin", "sfx_trared", "sfx_trago"
)

local function paVoiceClip(p)
	if (p.raceclock.timelimit - p.realtime) > p.raceclock.nextnotification*TICRATE then return end //Not due a notification
	local notification = (p.raceclock.timelimit - p.realtime) / 35
	if ((p.raceclock.timelimit - p.realtime) % 35 ~= 0) or p.raceclock.freezetime then return end
	if notification >=120 and notification <=122 then
		S_StartSound(nil,sfx_tra120,p)
		p.raceclock.nextnotification = 82
	elseif notification >=80 and notification <=82 then
		S_StartSound(nil,sfx_tra080,p)
		p.raceclock.nextnotification = 42
	elseif notification >=40 and notification <=42 then
		S_StartSound(nil,sfx_tra040,p)
		p.raceclock.nextnotification = 14
	elseif notification == 14 then
		S_StartSound(nil,sfx_tra010,p)
	elseif notification == 9 then
		S_StartSound(nil,sfx_tra009,p)
	elseif notification == 8 then
		S_StartSound(nil,sfx_tra008,p)
	elseif notification == 7 then
		S_StartSound(nil,sfx_tra007,p)
	elseif notification == 6 then
		S_StartSound(nil,sfx_tra006,p)
	elseif notification == 5 then
		S_StartSound(nil,sfx_tra005,p)
	elseif notification == 4 then
		S_StartSound(nil,sfx_tra004,p)
	elseif notification == 3 then
		S_StartSound(nil,sfx_tra003,p)
	elseif notification == 2 then
		S_StartSound(nil,sfx_tra002,p)
	elseif notification == 1 then
		S_StartSound(nil,sfx_tra001,p)
	end
end

// local function announcerVoiceClip(p)  // code shouldn't work, no clue how to setup the announcer within this lua - prof renderer
//	addHook("ThinkFrame", do
//		if (not (gametype & GT_MATCH)) and gpmode == true then
//			outplayers = 0

local function setupplayer(p)
	p.raceclock = {}
	p.raceclock.timelimit = 0 						//How much time is left in frames
	p.raceclock.startingtime = 0					//How much time did they start with?
	p.raceclock.freezetime = 0 						//How many frames to freeze the timer for
	p.raceclock.clearedmaps = 0 					//How many maps have the player cleared?
	p.raceclock.score = 0 							//The score which the player has
	p.raceclock.bonusscore = 0 						//Used to show score bonuses at the end of each lap/race
	p.raceclock.bonustype = ""						//A string which tells the current bonus type
	p.raceclock.bonusflash = 0 						//Used for the HUD element
	p.raceclock.oldtimelimit = 0 					//The old time limit so that time can be added to it
	p.raceclock.laps = 0 							//How many laps have they cleared?
	p.raceclock.priorlapstart = 0					//What time was left at the start of the last lap?
	p.raceclock.warningtext = {[1] = "", [2] = ""}	//Text displaying below the bonus regarding if you're in danger
	p.raceclock.oldspectate = p.spectator 			//Were they a spectator?
	p.raceclock.blewit = false 						//Have they blown the clock?
	p.raceclock.gonnablow = false 					//Are they about to blow the clock?
	p.raceclock.nextnotification = 0 				//When would they next get a timer update notification?
	p.raceclock.totalbonus = 0 						//Total bonus time obtained
	p.raceclock.lastmapcleared = nil 				//The last map the player cleared, if any
	p.raceclock.streak = 0 							//How many maps have you been gaining time in a row?
	p.raceclock.maxstreak = 0 						//Your best Streak
	p.raceclock.dwtime = 0							//Only used in Death Wish for how much time you should had gotten per map
end

local function getMapTime(map, p)
	local maptime
	
	if tonumber(mapheaderinfo[map].clocktime) ~= nil then //Custom map time?
		maptime = tonumber(mapheaderinfo[map].clocktime)
	elseif tonumber(mapheaderinfo[map].arcadetime) ~= nil and tonumber(mapheaderinfo[map].arcadebonus) ~= nil then //Is the map already Daytona Kart compatible? If so use that time.
		maptime = tonumber(mapheaderinfo[map].arcadetime) + ((tonumber(mapheaderinfo[map].arcadebonus) * (mapheaderinfo[map].numlaps - 1)))
	else //All other maps
		maptime = maptimes[map]
	end
	if maptime % 5 > 0 then //maptime is always a multiple of 5s, so round up if this is not the case.
		maptime = $ - ($ % 5) + 5
	end
	if maptime < 60 then maptime = 60
	elseif maptime > 300 then maptime = 300 end //Map times must be between 1 and 5 mins
	
	if dwon == 1 then
		p.raceclock.dwtime = maptime
		if p.raceclock.clearedmaps == 0 then maptime = 1000*TICRATE
		else 
			maptime = ($ * TICRATE * dwdiff)/20
		end
	else
		maptime = (((100 + playdiff)*($ * TICRATE))/100)
	end
	return maptime
end

local function deathwishclockdamage(p)
	local newtime = p.raceclock.timelimit
	local mo = p.mo
	if p.kartstuff[k_respawn] == 47 then //Respawning, so lose 3 seconds
		newtime = $ - (3*TICRATE)
	end
	if p.kartstuff[k_wipeoutslow]
	or p.kartstuff[k_spinouttimer] 
	or p.kartstuff[k_squishedtimer] then
		if not mo.zspinout
			mo.zspinout = true
			newtime = $ - (2*TICRATE)
		end
	else
		mo.zspinout = nil
	end
	if mo.eflags & MFE_JUSTBOUNCEDWALL
	and not p.powers[pw_flashing] then
		newtime = $ - (TICRATE)
	end
	
	if newtime <= p.realtime then
		newtime = p.realtime
	elseif newtime + TICRATE >= p.realtime and newtime ~= p.raceclock.timelimit then
		p.powers[pw_flashing] = TICRATE
		S_StartSound(mo, sfx_buzz2)
		K_DoInstashield(p)
	end
	if newtime ~= p.raceclock.timelimit then p.raceclock.timelimit = newtime end
end

local function settimelimit(map, p, newgame)
	if newgame or (not p.raceclock) then
		setupplayer(p)
	end
	if p.raceclock.oldspectate then
		p.raceclock.oldspectate = false
		
		//Reset the values to their default as this is a new run
		p.raceclock.oldtimelimit = 0 				
		p.raceclock.freezetime = 0 				
		p.raceclock.clearedmaps = 0 			
		p.raceclock.score = 0
		p.raceclock.blewit = false
	end
	if adaptiveDiff.value == 1 and p.raceclock.clearedmaps ~= 0 then
		if p.raceclock.streak > 0 and (p.raceclock.streak % 6) == 0 then
			if playdiff == NOOBMOD then playdiff = EASIESTMOD
			elseif playdiff == EASIESTMOD then playdiff = EASYMOD
			elseif playdiff == EASYMOD then playdiff = 0
			elseif playdiff == 0 then playdiff = HARDMOD
			elseif playdiff == HARDMOD then playdiff = EXPERTMOD
			elseif playdiff == EXPERTMOD then playdiff = MASTERMOD
			elseif playdiff == MASTERMOD then playdiff = INSANEMOD end
		end
	elseif p.raceclock.clearedmaps == 0 or forceUpdate.value == 1 then
		playdiff = difficulty.value
	end
	if (gpmode.value == 1) then gpmapsetup(p) end
	if dwon ~= deathwish.value then
		dwon = deathwish.value
		p.raceclock.oldtimelimit = 0 				
		p.raceclock.freezetime = 0 				
		p.raceclock.clearedmaps = 0 			
		p.raceclock.score = 0
		p.raceclock.blewit = false
	end
	if dwon and p.raceclock.clearedmaps == 0 then
		dwdiff = deathwishdiff.value
		COM_BufInsertText(server, "advancemap Next")
	end
	
	p.raceclock.totalbonus = $ + p.raceclock.oldtimelimit
	p.raceclock.timelimit = p.raceclock.oldtimelimit
	p.raceclock.timelimit = $ + getMapTime(map, p)
	if p.raceclock.freezetime then
		//If the timer was frozen when you cleared the last map, the remaining
		//frozen time is given to you at the start of the next map
		p.raceclock.timelimit = $ + p.raceclock.freezetime
		p.raceclock.freezetime = 0
	end
	p.raceclock.startingtime = p.raceclock.timelimit
	p.raceclock.priorlapstart = p.raceclock.startingtime
	p.raceclock.warningtext = {[1] = "", [2] = ""}
	p.raceclock.bonusscore = 0
	p.raceclock.bonustype = ""
	p.raceclock.bonusflash = 0
	p.raceclock.laps = 0
	p.raceclock.driftboost = 0
	p.raceclock.lastdriftcharge = 0
	p.realtime = 0
	p.raceclock.nextnotification = min(p.raceclock.timelimit/TICRATE, 122)
	if p.raceclock.nextnotification < 120 then p.raceclock.nextnotification = min(p.raceclock.timelimit/TICRATE, 82) end
	if p.raceclock.nextnotification < 80 then p.raceclock.nextnotification = min(p.raceclock.timelimit/TICRATE, 42) end
end

local commentary = {
	deathwish = {
		clear = {
			crit     = {"\x81".."Lucky. But next map, your soul'll be mine!"},
			low      = {"\x81".."Not bad. But next map, your soul'll be mine!"},
			standard = {"\x81".."You're doing swell, kid. But get ready..."},
			great    = {
				"\x81".."You're taking this contract quite well, kid.",
				"\x81".."Keep the good work up, kid."
			},
			amazing  = {"\x81".."Not bad - you have loads of time left, kid!"}
		},
		lap = {
			crit     = {"\x81".."Get ready to hand your soul over!"},
			bad      = {"\x81".."I'll be waiting for you to fail."},
			standard = {"\x81".."Make mistakes, and I'll collect."},
			great    = {"\x81".."Keep driving. You'll eventually lose."}
		}
	},
	normal = {
        clear = {
            crit     = {"\x87".."Hooo, that was close! Better watch yerself, lassie."},
            low      = {"\x82".."Not bad, not bad at all, lassie. Keep it up!"},
            standard = {"\x84".."Good on yer, lassie! Onto the next scene."},
            great    = {"\x84".."Hoo, yer ahead of schedule! Good work!"},
            amazing  = {"\x84".."Yer on fire! Better than any of my peck necks could do!"}
        },
        lap 
        = {
            crit     = {
				"\x85".."Hold onto yer hats, owls! This bomb's gonna blow!",
				"\x85".."Good thing you're not driving my train, peck neck!"
			},
            bad      = {
				"\x82".."Watch it, lassie! Yer gonna set off the bomb!",
				"\x82".."Drive faster, you no-good peck neck!"
			},
            standard = {
				"\x84".."Lookin' good. Don't make any mistakes.",
				"\x84".."Keep the skilled driving up, lassie!"
			},
            great    = {
				"\x83".."Good lap, lassie. We're still on schedule.",
				"\x83".."DJ Grooves should be embarrassed by that!"
			}
		}
	}
}

local function getcommentary(table)
	if not table then 
		//print("Couldn't find a table!")
		return ""
	end
	local entry = P_RandomRange(1, #table)
	if not table[entry] then return "" end
	//print(table[entry])
	return table[entry]
end

local function padTimerZero(num)
	if num < 10
		return "0" .. tostring(num)
	else
		return tostring(num)
	end
end

local function clearmap(p)
	local mapbonus
	local bonussecs = (p.raceclock.timelimit - p.realtime) / TICRATE
	local bonusframes = (p.raceclock.timelimit - p.realtime) % TICRATE
	local mod = 100
	local secmod = 5
	if dwon then
		mod = 10
		if dwdiff > 10 then 
			mod = $ - ((dwdiff - 10) / 2)
		else 
			mod = $ + (10 - dwdiff)
		end
	elseif playdiff < 0 then 
		mod = $ + (playdiff * (playdiff/5))
	elseif playdiff > 0 then 
		mod = $ - playdiff
		if playdiff == NOOBMOD then
			secmod = 1
		elseif playdiff == EASIESTMOD then
			secmod = 2
		else //playdiff == EASYMOD
			secmod = 3
		end
	end
	
	if (driftbonus.value == 1) then
		if not dwon then mod = $/2 --Because drifting gives you extra time, the bonus is cut in half
		else mod = ($*3)/4 end --Or by a quarter in Death Wish Mode
	end
	
	mapbonus = (bonussecs * (mod*secmod)) + (bonusframes * ((mod*12)/100))
	p.raceclock.score = $ + mapbonus
	p.raceclock.bonusscore = $ + mapbonus
	p.raceclock.bonustype = "MAP"
	p.raceclock.bonusflash = FLASHTIME
	p.raceclock.clearedmaps = $ + 1
	if p.raceclock.oldtimelimit < (p.raceclock.timelimit - p.realtime) then
		p.raceclock.streak = $ + 1 //Increase the streak!
		if p.raceclock.maxstreak < p.raceclock.streak then p.raceclock.maxstreak = p.raceclock.streak end
	elseif p.raceclock.oldtimelimit > (p.raceclock.timelimit - p.realtime) then
		p.raceclock.streak = 0  //Streak broken
	end
	p.raceclock.oldtimelimit = (p.raceclock.timelimit - p.realtime)
	
	local timeleft = p.raceclock.timelimit - p.realtime
	local warningstringtwo
	
	
	local commenttable
	
	if dwon then
		commenttable = commentary.deathwish.clear
	else
		commenttable = commentary.normal.clear
	end
	
	if timeleft <= 10*TICRATE then
		warningstringtwo = getcommentary(commenttable.crit)
	elseif timeleft <= 30*TICRATE then
		warningstringtwo = getcommentary(commenttable.low)
	elseif timeleft <= 90*TICRATE then
		warningstringtwo = getcommentary(commenttable.standard)
	elseif timeleft <= 180*TICRATE then
		warningstringtwo = getcommentary(commenttable.great)
	else
		warningstringtwo = getcommentary(commenttable.amazing)
	end
	
	if warningstringtwo then p.raceclock.warningtext[2] = warningstringtwo else p.raceclock.warningtext[2] = "" end
	
	local act = ""
	if mapheaderinfo[gamemap].actnum ~= 0
		act = " " .. tostring(mapheaderinfo[gamemap].actnum)
	end
	if (mapheaderinfo[gamemap].levelflags & LF_NOZONE) then
		p.raceclock.lastmapcleared = mapheaderinfo[gamemap].lvlttl .. act
	elseif mapheaderinfo[gamemap].zonttl ~= "" then
		p.raceclock.lastmapcleared = mapheaderinfo[gamemap].lvlttl .. " " .. mapheaderinfo[gamemap].zonttl .. act
	else
		p.raceclock.lastmapcleared = mapheaderinfo[gamemap].lvlttl .. " ZONE" .. act
	end
	if encoremode then p.raceclock.lastmapcleared = $ .. "E" end
end

local function clearlap(p)
	local mod = 100
	if dwon then mod = (400 - (dwdiff*10))/2
	elseif playdiff < 0 then mod = $ + (playdiff * (playdiff/5))
	elseif playdiff > 0 then mod = $ - playdiff end
	
	if (driftbonus.value == 1) then
		if not dwon then mod = $/2 --Because drifting gives you extra time, the bonus is cut in half
		else mod = ($*3)/4 end --Or by a quarter in Death Wish Mode
	end
	
	p.raceclock.score = $ + (mod * p.laps)
	p.raceclock.bonusscore = (mod * p.laps)
	p.raceclock.bonustype = "LAP"
	p.raceclock.bonusflash = FLASHTIME
	p.raceclock.laps = $ + 1
	
	local timeleft = p.raceclock.timelimit - p.realtime
	local lastlaptime = p.raceclock.priorlapstart - timeleft
	p.raceclock.priorlapstart = timeleft
	local lattimestring = padTimerZero(G_TicsToMinutes(lastlaptime, true)) .. "'" .. padTimerZero(G_TicsToSeconds(lastlaptime)) .. "\"" ..  padTimerZero(G_TicsToCentiseconds(lastlaptime))
	local expectedtimeperlap = p.raceclock.startingtime / mapheaderinfo[gamemap].numlaps
	if dwon then
		expectedtimeperlap = p.raceclock.dwtime / mapheaderinfo[gamemap].numlaps
	end
	local warningstringone
	local warningstringtwo
	if lastlaptime > (expectedtimeperlap + (5*TICRATE)) then
		warningstringone = "\x85".."Bad Lap: ".."\x80"..lattimestring
	elseif lastlaptime > expectedtimeperlap then
		warningstringone = "\x82".."Slow Lap: ".."\x80"..lattimestring
	elseif (lastlaptime + (5*TICRATE)) >= expectedtimeperlap then
		warningstringone = "\x84".."Good Lap: ".."\x80"..lattimestring
	else
		warningstringone = "\x83".."Fast Lap: ".."\x80"..lattimestring
	end
	if p.raceclock.laps < mapheaderinfo[gamemap].numlaps then
	
		local commenttable
	
		if dwon then
			commenttable = commentary.deathwish.lap
		else
			commenttable = commentary.normal.lap
		end
	
		local averagetimeperlap = timeleft / (p.raceclock.laps - mapheaderinfo[gamemap].numlaps)
		if averagetimeperlap >= (expectedtimeperlap + (5*TICRATE)) 
			warningstringtwo = getcommentary(commenttable.crit)
		elseif averagetimeperlap > expectedtimeperlap then 
			warningstringtwo = getcommentary(commenttable.bad)
		elseif (averagetimeperlap + (5*TICRATE)) >= expectedtimeperlap then 
			warningstringtwo = getcommentary(commenttable.standard)
		else 
			warningstringtwo = getcommentary(commenttable.great)
		end
	end
	p.raceclock.warningtext[1] = warningstringone
	if warningstringtwo then p.raceclock.warningtext[2] = warningstringtwo else p.raceclock.warningtext[2] = "" end
end

local function blowtheclock(p)
	if not (p and p.mo) or p.spectator return end
	if p.blown return end
	p.pflags = $|PF_TIMEOVER
	p.lives = 0
	p.blown = true
	P_DamageMobj(p.mo, nil, nil, 10000)
	p.blown = nil
	P_FlashPal(p, 1, 10)
	S_ChangeMusic("KRFAIL", true, p)
	endtimer = 1
	p.raceclock.blewit = true
end

local function checkclockable()
	//This checks that there's only the player, and it's not a Record Attack
	if splitscreen or modeattacking then 
		clockrace = false
		return
	end
	
	local numplayers = 0
	for p in players.iterate do
		if p.spectator then 
			if p == players[clockattacker] then
				local nextplay = nextplayer(p)
				if nextplay == p then
					p.spectator = false
				else
					clockattacker = #nextplay
					if #nextplay < #p then numplayers = $ + 1 end
					continue
				end
			else
				continue
			end
		end //Allow spectators to watch
		numplayers = $ + 1
		if numplayers > 1 then 
			clockrace = false
			return
		end
		clockattacker = #p
	end
	
	if numplayers == 0 then
		clockrace = false
		return
	end //Can't Clock Attack with no players!
	
	clockrace = true //All checks passed, so it's clockable!
end

local function drawLetter(v, font, char, x, y, f, c)
	local asc = string.byte(char)
	local padding = "000"
	if asc > 99
		padding = ""
	elseif asc > 9
		padding = "0"
	elseif asc > 0
		padding = "00"
	end
	v.draw(x, y, v.cachePatch(font .. padding .. tostring(asc)), f, c)
end

local function drawLetterScaled(v, font, char, x, y, scale, f, c)
	local asc = string.byte(char)
	local padding = "000"
	if asc > 99
		padding = ""
	elseif asc > 9
		padding = "0"
	elseif asc > 0
		padding = "00"
	end
	v.drawScaled(x, y, scale, v.cachePatch(font .. padding .. tostring(asc)), f, c)
end

local function drawFontString(v, font, str, x, y, spacing, f, c)
	for a=1,string.len(str) do
		drawLetter(v, font, string.sub(str, a, a), x + ((a - 1) * spacing), y, f, c)
	end
end

local function drawScaledFontString(v, font, str, x, y, spacing, scale, f, c)
	for a=1,string.len(str) do
		drawLetterScaled(v, font, string.sub(str, a, a), x + ((a - 1) * spacing), y, scale, f, c)
	end
end

local function drawMKtext(v, str, x, y, flags)
	if str == nil
	or string.len(str) == 0
		str = "NIL"
	end
	str = tostring(str)
	str = str:upper()
	drawFontString(v, "MKFNT", str, x, y, 12, flags)
end

local function drawScore(v, p, flags)
	local scorestring = tostring(p.raceclock.score)
	local stringlen = string.len(scorestring)
	while stringlen < 7 do
		scorestring = "0" .. $
		stringlen = $ + 1
	end
	
	v.drawString(160, 2, "SCORE:", V_BLUEMAP|flags, "center")
	local xpos = 160 - (6 * stringlen)
	drawMKtext(v, scorestring, xpos, 12, flags)
	local difficultyText = ({
		[INSANEMOD]  = "Insane",
		[MASTERMOD]  = "Master",
		[EXPERTMOD]  = "Expert",
		[HARDMOD]    = "Hard",
		[0]          = "Normal",
		[EASYMOD]    = "Easy",
		[EASIESTMOD] = "Easiest",
		[NOOBMOD]    = "N00B"
	}) [playdiff]
	local difficultyFlags = (playdiff == NOOBMOD and V_GRAYMAP) or (playdiff > 0 and V_BLUEMAP) or (((playdiff == MASTERMOD) or (playdiff == INSANEMOD)) and V_REDMAP) or (playdiff < 0 and V_ORANGEMAP) or V_YELLOWMAP
	if (driftbonus.value == 1) then 
		difficultyText = $.." [DTF ON]"
	end
	if dwon then
		local dwstring = "DEATH WISH"
		if dwdiff > 10 then
			dwstring = "MINOR "..$
		elseif dwdiff == 0 then
			dwstring = "REAL "..$
		elseif dwdiff < 10 then
			dwstring = "MAJOR"..$
		end
		v.drawString(160, 28, dwstring, flags|V_PURPLEMAP, "center")
	elseif gpmode.value then
		v.drawString(160, 28, "Race " .. (p.raceclock.clearedmaps + (p.laps ~= numlaps and 1 or 0)) .. ": " .. difficultyText, flags|difficultyFlags, "center")
	else
		v.drawString(160, 28, "Level: " .. difficultyText, flags|difficultyFlags, "center")
	end
	
	if p.raceclock.bonusflash > 0 then
		local fontflash = (p.raceclock.bonusflash % 30) / 3
		if fontflash > 5 then fontflash = 10 - $ end //This gives a number between 0 and 5 which pulses
		fontflash = $ * 2
		local flashflag = flash[fontflash]
		v.drawString(160, 38, "+" .. tostring(p.raceclock.bonusscore) .. ": " .. p.raceclock.bonustype .. " CLEAR!", V_GREENMAP|V_SNAPTOTOP|flashflag, "center")
		
		if comments.value == 1 then
			if p.raceclock.warningtext[1] and (p.raceclock.warningtext[1] ~= "") then
				v.drawString(160, 48, p.raceclock.warningtext[1], V_SNAPTOTOP|flashflag, "center")
			end
			if p.raceclock.warningtext[2] and (p.raceclock.warningtext[2] ~= "") then
				local xpos = 160 - (2*string.len(p.raceclock.warningtext[2]))
				v.drawString(xpos, 58, p.raceclock.warningtext[2], V_SNAPTOTOP|flashflag, "small")
			end
		end
	end
end

local function padTimerZero(num)
	if num < 10
		return "0" .. tostring(num)
	else
		return tostring(num)
	end
end

local TIMEUPDATEFIN = 16 + (TICRATE * 6)
local TIMEUPDATESTART = 26 + (TICRATE * 3)
local UPDATEMOD = TIMEUPDATEFIN - TIMEUPDATESTART
local MAXTIME = 2000
//local timepatch = CV_RegisterVar({"rtc_patchclock", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})

local function drawtimer(v, p, flags)
	local timer = (p.raceclock.timelimit - p.realtime)
	local addtime = false
	if (leveltime < (TIMEUPDATEFIN)) then
		addtime = true
		timer = max(p.raceclock.oldtimelimit, p.raceclock.timelimit - max(((maptimes[map] * TICRATE)/UPDATEMOD) * (TIMEUPDATEFIN - leveltime), 0))
	end
	if timer < 0 then timer = 0 end
	local overtime = false
	local timecap = MAXTIME
	if timer >= timecap*TICRATE then //The clock can only show up to 199.97 seconds
		timer = (timecap*TICRATE)
		overtime = true
	end
	local timerpatch = v.cachePatch("HITTMER")
	if dwon then 
		if dwsoul.value then
			timerpatch = v.cachePatch("DWSTMER")
		else
			timerpatch = v.cachePatch("DWTMER")
		end
	end
	local timerpatchscale = FRACUNIT
	
	local pulse = 21 - ((p.realtime - (p.raceclock.timelimit % 35)) % 35)
	if addtime then pulse = 0 end
	if (((timer / TICRATE) % 10) > 1) and (((timer / TICRATE) % 10) < 9) and ((timer/TICRATE) > 10) then pulse = 0 end
	if pulse == 21 then pulse = 0 end
	if pulse < 0 then pulse = 0 end //This gives a value between 0 and 20
	timerpatchscale = $ + (pulse * FRACUNIT / 50)
	
	local timersec = padTimerZero(timer / TICRATE)
	local timercenti = padTimerZero(G_TicsToCentiseconds(timer))
	if timer == timecap*TICRATE then
		timersec = tostring(timecap - 1)
		timercenti = "97"
	end
	
	local fontstring = ((addtime and "TMERG") or (p.raceclock.freezetime and "TMER") or (overtime and "TMERG") or ((((timer/TICRATE) < 10) and (pulse > 5) or (timer == 0)) and "TMERR") or "TMERY")
	
	v.drawScaled(320*FRACUNIT, 200*FRACUNIT, timerpatchscale, timerpatch, flags)
	
	local secscale = timerpatchscale / 2
	local xpos = -71
	local ypos = -60
	if string.len(timersec) == 3 then xpos = -83
	elseif string.len(timersec) == 4 then xpos = -99 end
	
	local secx = 320*FRACUNIT + (xpos * timerpatchscale / 2)
	local secy = 200*FRACUNIT + (ypos * timerpatchscale / 2)
	drawScaledFontString(v, fontstring, timersec, secx, secy, 17*secscale, secscale, flags)
	
	local centiscale = timerpatchscale / 4
	xpos = -15
	ypos = -23
	local centix = 320*FRACUNIT + (xpos * timerpatchscale)
	local centiy = 200*FRACUNIT + (ypos * timerpatchscale)
	drawScaledFontString(v, fontstring, timercenti, centix, centiy, 17*centiscale, centiscale, flags)
end


local function forcespectate(p)
	if p ~= players[clockattacker] then //Don't force the active player to spectate
		p.spectator = true
	end
end


local function nextplayer(oldplayer)
	//Returns the player number of the next active player, and sets the old active player to spectate
	local firstplayer = nil
	local lastplayer
	local getplayer = false
	
	for p in players.iterate do
		if p.valid then
			if firstplayer == nil then firstplayer = #p end
			if getplayer then return #p end
			if (not p.spectator) or (p == oldplayer) then
				getplayer = true
				p.spectator = true
				p.raceclock.oldspectate = true
			end
		end
	end
	
	return firstplayer
end

local function deathseq(p)
	local mo = p.mo
	//mo.zhudshake = 69	-- force hud to shake so that it looks cool
	mo.zspinoutdeath = $-1
	p.realtime = p.raceclock.timelimit

	if not (mo.zspinoutdeath or p.exiting) then
		blowtheclock(p)
	end
end

addHook("MapLoad", do
	endtimer = 0 //Reset endtime right away
	if not (gametype & GT_MATCH) then
		checkclockable()
		if clockrace then
			local activeplayer = players[clockattacker]
			settimelimit(gamemap, activeplayer)
			if activeplayer.raceclock.blewit or activeplayer.raceclock.oldspectate then
				setupplayer(activeplayer)
			end
			for p in players.iterate do
				if not p.raceclock then setupplayer(p) end
				p.raceclock.gonnablow = false
				p.raceclock.blewit = false
				p.raceclock.finann = false
			end
		else
			print("Unable to race the clock!")
		end
	end
end)

//Lock-on mod for Acrobatics which allows for time to freeze for a short while on a successful trick.
//Death Wish will cause damage to the timer on a failed trick though
rawset(_G, "kartrush_acrolock", function(p, tricktype)
	if p.trickgrade <= 0 then
		if dwon then 
			local newtime = p.raceclock.timelimit - TICRATE
			if newtime <= p.realtime then
				newtime = p.realtime
			elseif newtime + TICRATE >= p.realtime and newtime ~= p.raceclock.timelimit then
				p.powers[pw_flashing] = TICRATE
				S_StartSound(mo, sfx_buzz2)
				K_DoInstashield(p)
			end
			if newtime ~= p.raceclock.timelimit then p.raceclock.timelimit = newtime end
		end
	else
		if tricktype == 1 then
			p.raceclock.freezetime = $ + (p.trickgrade * (TICRATE/2))
		elseif tricktype == 0 then
			p.raceclock.freezetime = $ + (TICRATE/2)
		end
	end
end)

addHook("ThinkFrame", do
	if (not (gametype & GT_MATCH)) and clockrace then
		local p = players[clockattacker]
		
		if not p then return end
		if not p.raceclock then return end //Wait until MapLoad has set the time limit.
		
		p.kartstuff[k_lapanimation] = 0
		
		if (not p.raceclock.oldspectate) and p.spectate then 
			p.raceclock.oldspectate = true
			local nxtplr = nextplayer(players[clockattacker])
			clockattacker = #nxtplr
			p = players[clockattacker]
			p.spectator = false
			if dwon == 1 then
				COM_BufInsertText(server, "map map01")
			else
				COM_BufInsertText(server, "advancemap Same")
				G_ExitLevel()
			end
		end
		
		if p.exiting and (not p.raceclock.finann) then
			p.raceclock.finann = true
			S_StartSound(nil,sfx_trafin,p)
		end
		
		if dwon and leveltime >= TIMEFORGO then
			deathwishclockdamage(p)
		end
		
		if (leveltime == 0) then
			if (gpmode.value == 1) and (gpmusic.value == 1) and (playdiff == INSANEMOD) then
				S_ChangeMusic("MULINE", false, p)
			elseif dwon then
				if p.raceclock.clearedmaps == 0 then
					S_ChangeMusic("DWSTRT", false, p)
				else
					S_ChangeMusic("DWANTI", false, p)
				end
			end
		end
		
		
		if leveltime == TIMEFORTHREE then
			S_StartSound(nil,sfx_trared,p)
		elseif leveltime == TIMEFORGO then
			S_StartSound(nil,sfx_trago,p)
		end
		
		if (leveltime == TIMEFORGO + (TICRATE/2) - 1) then
			if dwon == 1 then
				COM_BufInsertText(server, "tunes DEAWIS") // Plays Death Wish when Death Wish Mode is on - Professor Renderer
			elseif (gpmode.value == 1) and (gpmusic.value == 1) then
				if encoremode then
					COM_BufInsertText(server, "tunes CRIFT")
				elseif (playdiff == EASIESTMOD) or (playdiff == NOOBMOD) then
					COM_BufInsertText(server, "tunes CALMHR") // Plays Calm Hour when GP Mode is first started - Professor Renderer
				elseif playdiff == EASYMOD then
					COM_BufInsertText(server, "tunes RUSHHR") // Plays Rush Hour when GP Mode difficulty is Easy - Professor Renderer
				elseif playdiff == HARDMOD then
					COM_BufInsertText(server, "tunes HARTRN") // Plays sped up Train Rush when GP Mode difficulty is Hard - Professor Renderer
				elseif playdiff == EXPERTMOD then
					COM_BufInsertText(server, "tunes WUPWND") // Plays Wound-up Windmill when GP Mode difficulty is Expert - Professor Renderer
				elseif playdiff == MASTERMOD then
					COM_BufInsertText(server, "tunes YAABGS") // Plays You Are All Bad Guys when GP Mode difficulty is Master - Professor Renderer
				elseif playdiff == INSANEMOD then
					COM_BufInsertText(server, "tunes MUSTEX") // Plays Mustache Girl EX when GP Mode difficulty is Insane - Professor Renderer
				else
					COM_BufInsertText(server, "tunes TRNRSH") // Plays Train Rush when GP Mode difficulty is Normal - Professor Renderer
				end
			end
		end
		
		if p.raceclock.laps < p.laps then
			clearlap(p)
			if p.laps == numlaps then
				clearmap(p)
			end
			p.score = p.raceclock.score
		end
		
		if p.raceclock.bonusflash then
			p.raceclock.bonusflash = $ - 1
		end
		
		if driftbonus.value == 1 then
			if p.kartstuff[k_driftboost] > p.raceclock.driftboost then
				if p.raceclock.lastdriftcharge >= K_GetKartDriftSparkValue(p)*4 then
					p.raceclock.freezetime = $ + (TICRATE*10/3) + 1
				elseif p.raceclock.lastdriftcharge >= K_GetKartDriftSparkValue(p)*2 then
					p.raceclock.freezetime = $ + (TICRATE*3/2) + 1
				elseif p.raceclock.lastdriftcharge >= K_GetKartDriftSparkValue(p) then
					p.raceclock.freezetime = $ + (TICRATE*2/3) + 1
				end
				p.raceclock.driftboost = p.kartstuff[k_driftboost]
			elseif p.kartstuff[k_driftboost] != p.raceclock.driftboost then
				p.raceclock.driftboost = p.kartstuff[k_driftboost]
			end
			p.raceclock.lastdriftcharge = p.kartstuff[k_driftcharge]
		end
		
		if p.raceclock.freezetime and (not p.exiting) then
			p.raceclock.timelimit = $ + 1
			p.raceclock.freezetime = $ - 1
		end
		
		if p.mo and p.mo.valid and p.mo.zspinoutdeath then 
			deathseq(p)
			return
		end
		
		if p.raceclock.timelimit == p.realtime and (not p.raceclock.gonnablow) then
			local mo = p.mo
			p.kartstuff[k_spinouttimer] = TICRATE*8
			p.kartstuff[k_sneakertimer] = 0
			mo.zspinoutdeath = TICRATE
			P_DamageMobj(mo)
			P_PlayRinglossSound(mo, p)
			if p == displayplayer
				P_StartQuake(60<<FRACBITS, TICRATE*2)
			end
			p.raceclock.gonnablow = true
			if trainPA.value == 1 then S_StartSound(nil,sfx_tra000,p) end
			return
		end
		
		if ((p.raceclock.timelimit - p.realtime) % 35 == 0) and (p.raceclock.freezetime == 0)
		and (not p.raceclock.gonnablow) and (trainPA.value == 1) and leveltime >= TIMEFORGO
		and ((p.raceclock.timelimit - p.realtime) <= p.raceclock.nextnotification*TICRATE) then
			paVoiceClip(p)
		end
		
		if endtimer > 0 then
			endtimer = $ + 1
			if endtimer >= TICRATE*8 and p.cmd.buttons & BT_CUSTOM1 then
				local nxtplr = nextplayer(players[clockattacker])
				clockattacker = nxtplr
				players[clockattacker].spectator = false
				G_ExitLevel()
			end
		end
	end
end)

local basemaxspeed = {
	2519263, //Speed 1
	2568660, //Speed 2
	2618057, //Speed 3
	2667455, //Speed 4
	2716852, //Speed 5
	2766249, //Speed 6
	2815647, //Speed 7
	2865044, //Speed 8
	2914441  //Speed 9
}

local function convertspeed(speed, speedtype, mobjscale)
	local convertedspeed
	local fractionspeed
	local typetext
	local result
	if speedtype == MILESPERHOUR then
		typetext = "MPH"
		convertedspeed = FixedDiv(FixedMul(speed, 88465), mobjscale)/(FRACUNIT/10)
	elseif speedtype == KILOMETERSPERHOUR then
		typetext = "KM/H"
		convertedspeed = FixedDiv(FixedMul(speed, 142371), mobjscale)/(FRACUNIT/10)
	elseif speedtype == METERSPERSECOND then
		typetext = "M/S"
		convertedspeed = FixedDiv(FixedMul(speed, 39547), mobjscale)/(FRACUNIT/10)
	elseif speedtype == FRACUNITSPERTIC then
		typetext = "FU/T"
		convertedspeed = FixedDiv(speed, mobjscale)/(FRACUNIT/10)
	elseif speedtype == NAUTICALMILESPERHOUR then
		typetext = "KNOT"
		convertedspeed = FixedDiv(FixedMul(speed, 76874), mobjscale)/(FRACUNIT/10)
	elseif speedtype == FOOTPERSECOND then
		typetext = "FT/S"
		convertedspeed = FixedDiv(FixedMul(speed, 129747), mobjscale)/(FRACUNIT/10)
	end
	fractionspeed = convertedspeed % 10
	convertedspeed = $ / 10
	return convertedspeed, fractionspeed, typetext
end

local RPMTIC = 125
local BASERPM = 8000

local function boostchargerpm(charge, perfectRPM, overRPM)
	//Returns 1000 if charge == 0, approx perfectRPM if charge is 35 or 36,
	//slightly over if 37-40, and at least overRPM if 50 or more
	
	local MINRPM = 1000
	local phaseOne = (perfectRPM - MINRPM)/35
	local phaseTwo = perfectRPM/80
	local phaseThree = (overRPM - (perfectRPM + (perfectRPM/20)))/10
	local phaseMax = overRPM/200
	
	local result = MINRPM
	if charge <=35 then
		result = $ + phaseOne*charge
		return result
	end
	result = perfectRPM
	if charge == 36 then return result end
	if charge <= 40 then
		result = $ + phaseTwo*(charge-36)
	elseif charge <= 49 then
		result = $ + phaseTwo*4 + phaseThree*(charge-40)
	else
		result = overRPM + phaseMax*(charge-50)
	end
	return result
end

local function drawspeed(v, p, x, y, flags)
	local rpm = 1000 + FixedDiv(FixedMul(FixedDiv(p.speed, mapheaderinfo[gamemap].mobj_scale), (BASERPM - 1000)), basemaxspeed[p.kartspeed])
	//Do custom RPM checks for during the start of a race here
	if leveltime <= TIMEFORGO then
		rpm = boostchargerpm(p.kartstuff[k_boostcharge], BASERPM, 10000)
	end
	//x = 18
	//y = 150
	
	local rpmtics = (rpm/RPMTIC) + 1
	local wholespeed
	local tenthspeed
	local typetext
	local speedtext
	wholespeed, tenthspeed, typetext = convertspeed(p.speed, customspeedometer.value, mapheaderinfo[gamemap].mobj_scale)
	
	//Draw base image
	v.draw(x, y, v.cachePatch("HUDSPEED"), flags)
	
	if rpmtics > 7 then
		if rpmtics < 32 then
			v.drawFill(x+10, y+10, (rpmtics - 8), 5, 187)
			local transparency = (rpm%RPMTIC)/(RPMTIC/10)
			v.draw(x + 2 + rpmtics, y+10, v.cachePatch("HUDSPD1"), flash[transparency])
		else
			v.drawFill(x+10, y+10, 24, 5, 187)
			if rpmtics < 64 then
				v.drawFill(x+34, y+10, (rpmtics - 32), 7, 106)
				local transparency = (rpm%RPMTIC)/(RPMTIC/10)
				v.draw(x + 2 + rpmtics, y+10, v.cachePatch("HUDSPD2"), flash[transparency])
			else
				v.drawFill(x+34, y+10, 32, 7, 106)
				if rpmtics > 87 then
					v.drawFill(x+66, y+10, 24, 11, 133)
				else
					v.drawFill(x+66, y+10, (rpmtics - 64), 11, 133)
					local transparency = (rpm%RPMTIC)/(RPMTIC/10)
					v.draw(x + 2 + rpmtics, y+10, v.cachePatch("HUDSPD3"), flash[transparency])
				end
			end
		end
	end
	
	//Draw Speed
	if tenthspeedometer.value == 0 then
		if tenthspeed > 4 then wholespeed = $ + 1 end
		speedtext = tostring(wholespeed)
	else
		speedtext = wholespeed .. "." .. tenthspeed
	end
	local speedx = 62 + x - (string.len(speedtext) * 12)
	drawMKtext(v, speedtext, speedx, y+22, flags)
	v.drawString(x+64, y+28, typetext, flags, "left")
end

local valuenames = {
	"BASE SCORE",
	"MAPS CLEARED",
	"BONUS SECS",
	"BEST STREAK"
}

local function clockResults(v, p)
	if splitscreen return end
	if not p.mo return end
	
	local timer = endtimer
	if not timer return end	-- uh what

	if timer > TICRATE*2
		local fade = v.cachePatch("BLACKSQR")
		local t = max(V_50TRANS, V_90TRANS - V_10TRANS*((min(50, timer-TICRATE*2))/4))
		v.draw(0, 0, fade, V_SNAPTOTOP|V_SNAPTOLEFT|t)
	end
	
	local ypos = 40

	if timer > TICRATE*4
		if dwon == 1 then
			drawMKtext(v, "DEATH WISH RESULTS", 52, ypos, V_SNAPTOTOP)
		elseif gpmode.value == 1 then
			drawMKtext(v, "GP RESULTS", 100, ypos, V_SNAPTOTOP)
		else
			drawMKtext(v, "RESULTS", 118, ypos, V_SNAPTOTOP)
		end
	end
	
	local vals = {
		p.raceclock.score,
		p.raceclock.clearedmaps,
		p.raceclock.totalbonus/TICRATE,
		p.raceclock.maxstreak
	}
	
	local mods = {
		1,
		0,
		((dwon == 1) and (300 - (dwdiff*5)) or (gpmode.value == 1) and (min(60 + (p.raceclock.clearedmaps * 4), 200)) or 100 - playdiff),
		(((dwon == 1) and -1) or (gpmode.value == 1) and (min(500 + (p.raceclock.clearedmaps * 5), 2500)) or 1000 - (abs(playdiff) * (playdiff/5) * 5))
	}
	
	ypos = $ + 20
	
	if timer > TICRATE*4 + (TICRATE/3) else return end
	v.drawString(50, ypos, "LAST CLEARED MAP:", V_YELLOWMAP)
	if p.raceclock.lastmapcleared ~= nil
		v.drawString(260, ypos + 10, p.raceclock.lastmapcleared, V_BLUEMAP, "right")
	else
		v.drawString(260, ypos + 10, "NONE", V_REDMAP, "right")
	end
	
	local fatalMap
	
	local act = ""
	if mapheaderinfo[gamemap].actnum ~= 0
		act = " " .. tostring(mapheaderinfo[gamemap].actnum)
	end
	if (mapheaderinfo[gamemap].levelflags & LF_NOZONE) then
		fatalMap = mapheaderinfo[gamemap].lvlttl .. act
	elseif mapheaderinfo[gamemap].zonttl ~= "" then
		fatalMap = mapheaderinfo[gamemap].lvlttl .. " " .. mapheaderinfo[gamemap].zonttl .. act
	else
		fatalMap = mapheaderinfo[gamemap].lvlttl .. " ZONE" .. act
	end
	if encoremode then fatalMap = $ .. "E" end
	
	v.drawString(50, ypos + 20, "FATAL MAP:", V_YELLOWMAP)
	v.drawString(260, ypos + 30, fatalMap, V_REDMAP, "right")
	
	for i = 1, #vals do
		if timer > TICRATE*5 + (TICRATE/3*(i)) else break end
		if mods[i] > 0 then
			v.drawString(60, ypos+40+i*10, valuenames[i], V_YELLOWMAP) -- Altered to take into account the extra lines
			v.drawString(250, ypos+40+i*10, vals[i], 0, "right")
			if mods[i] > 1
				v.drawString(160, ypos+40+i*10, mods[i].."x", V_YELLOWMAP)
			end
		elseif mods[i] == 0 then
			v.drawString(60, ypos+40+i*10, valuenames[i], V_YELLOWMAP)
			v.drawString(250, ypos+40+i*10, vals[i], V_GREENMAP, "right")
		end
	end
	
	if timer > TICRATE*6 + ((TICRATE/3)*2) else return end -- Altered to work with the extra values

	v.drawFill(60,  ypos + 95, 190, 1, 0) -- Altered to work with the extra values


	if timer > TICRATE*7 + (2*TICRATE/3) else return end -- Added
	local sum = 0
	for i = 1, #vals
		if mods[i] > 0 then
			sum = $ + vals[i]*mods[i]
		end
	end
	local timersum = (timer-(TICRATE*7+(2*TICRATE/3)))*50000
	
	v.drawString(60, ypos + 100, "TOTAL:", V_YELLOWMAP) -- Altered to work with the extra values
	if timersum >= sum then
		v.drawString(250, ypos + 100, sum, V_GREENMAP, "right")
		if not ktching
			S_StartSound(nil, sfx_chchng, p)
			ktching = true
		end
	else
		v.drawString(250, ypos + 100, timersum, 0, "right") -- Altered to work with the extra values
		S_StartSound(nil, sfx_menu1, p)
		ktching = false
	end
	
	if timer > TICRATE*8
		local waitstr
		if p.exiting
			waitstr = "EXITING..."
		else
			waitstr = "PRESS CUSTOM 1 TO CONFIRM"
		end
		v.drawString(160, 180, waitstr, V_SNAPTOBOTTOM|V_GREENMAP, "center")
	end
end

local function timeriftflash(v)
	local flashnum = (leveltime % (NUMSKINCOLORS*10))/10
	local colormap = v.getColormap(0, flashnum)
	local flags = V_50TRANS + (V_10TRANS * abs(4 - (leveltime % 32)/4))
	
	v.draw(0, 0, v.cachePatch("CTRSQR"), flags, colormap)
end

hud.add(function(v, p)
	if not (gametype & GT_MATCH) and clockrace and (p == players[clockattacker]) then
		if hud.enabled("time") then hud.disable("time") end
		if hud.enabled("position") then hud.disable("position") end
		if hud.enabled("freeplay") then hud.disable("freeplay") end
		if p.realtime < 1 then
			drawtimer(v, p, V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_HUDTRANS)
			drawScore(v, p, V_SNAPTOTOP|V_HUDTRANS)
		else
			drawtimer(v, p, V_SNAPTOBOTTOM|V_SNAPTORIGHT)
			drawScore(v, p, V_SNAPTOTOP)
		end
		if endtimer > 0 then 
			if hud.enabled("item") then hud.disable("item") end
			if hud.enabled("speedometer") then hud.disable("speedometer") end
			if hud.enabled("gametypeinfo") then hud.disable("gametypeinfo") end
			if hud.enabled("minimap") then hud.disable("minimap") end
			clockResults(v, p)
		else
			if not hud.enabled("item") then hud.enable("item") end
			if not hud.enabled("minimap") then hud.enable("minimap") end
			if dspeed.value == 1 then
				if hud.enabled("speedometer") then hud.disable("speedometer") end
				if hud.enabled("gametypeinfo") then hud.disable("gametypeinfo") end
				drawspeed(v, p, 4, 150, V_HUDTRANS)
				v.drawString(4, 190, "LAP " .. (p.laps < numlaps and (p.laps + 1) or p.laps) .. " OF " .. numlaps, V_SNAPTOBOTTOM|V_HUDTRANS)
			else
				if not hud.enabled("speedometer") then hud.enable("speedometer") end
				if not hud.enabled("gametypeinfo") then hud.enable("gametypeinfo") end
			end
			if encoremode and gpmode.value == 1 and gprift.value == 1 then
				timeriftflash(v)
			end
		end
	else
		if not hud.enabled("time") then hud.enable("time") end
		if not hud.enabled("position") then hud.enable("position") end
		if not hud.enabled("freeplay") then hud.enable("freeplay") end
		if not hud.enabled("item") then hud.enable("item") end
		if not hud.enabled("speedometer") then hud.enable("speedometer") end
		if not hud.enabled("gametypeinfo") then hud.enable("gametypeinfo") end
		if not hud.enabled("minimap") then hud.enable("minimap") end
	end
end)


hud.add(function(v)
	v.drawString(320, 192, "Kart Rush: "..MODVERSION.." ", V_ALLOWLOWERCASE|V_40TRANS, "right")
end, "scores")

print("\x82".."Kart Rush: Race the Clock version "..MODVERSION.." is now installed!")
print("\x82".."Use 'kr_welcome' to see the welcome text!")